import {Bus} from "./global";
import {connect, Connection} from "amqplib";
import Bluebird = require("bluebird");



export class RabbitMqBus implements Bus {

    connection: Promise<Connection>;
    private url: string;
    private connectionRetry: number;
    private readonly maxConnectionRetry: number;
    constructor({url, maxConnectionRetry}: { url: string, maxConnectionRetry?: number }) {
        this.url = url;
        this.maxConnectionRetry = maxConnectionRetry || 5;
        this.connectionRetry = 0;
        this.connection = this.connect();

    }

    connect(): Promise<Connection>{
        this.connectionRetry = this.connectionRetry + 1;
        return new Promise((resolve, reject) => {

            connect(this.url).then(connection => {
                resolve(connection)
            }).catch(err => {
                if(this.connectionRetry > this.maxConnectionRetry){
                    process.exit(500);
                }
                console.log('connection failed try #' + this.connectionRetry);
                setTimeout(() => {resolve(this.connect())}, 500)
            })
        })
    }


    private async channel(commandName: string) {
        const conn = await this.connection;
        const ch = await conn.createChannel();
        await ch.assertQueue(commandName);
        return ch;
    }

    async command(commandName: string, payload: any): Promise<void> {

        const ch = await this.channel(commandName);
        ch.sendToQueue(commandName, Buffer.from(JSON.stringify(payload)));
        await ch.close();
    }

    async listen(commandName: string, callback: (payload: any) => Promise<void>) {
        const ch = await this.channel(commandName);
        await ch.consume(commandName, async msg => {
            if (msg !== null) {
                ch.ack(msg);
                const payload = JSON.parse(msg.content.toString());
                await callback(payload);
            } else {
                ch.reject(msg)
            }
        });
    }

    async publish(eventName: string, payload: any): Promise<void> {
        const {ch, ok} = await this.exchange(eventName);
        ch.publish(eventName, '', Buffer.from(JSON.stringify(payload)));
        await ch.close()
    }

    async on(eventName: string, callback: (payload: any) => Promise<void>) {
        const {ch, ok} = await this.exchange(eventName);
        const {queue} = await ch.assertQueue('');
        await ch.bindQueue(queue, eventName, '');
        await ch.consume(queue,  async msg => {
            const payload = JSON.parse(msg.content.toString());
            await callback(payload);
        }, {noAck: true})
    }

    private async exchange(eventName: string) {
        const conn = await this.connection;
        const ch = await conn.createChannel();
        const ok = await ch.assertExchange(eventName, 'fanout', {durable: false});
        return {ch, ok};
    }
}